from abc import ABCMeta, abstractmethod

import numpy as np
from abcpy.distributions import Distribution

class Model(metaclass = ABCMeta):
    """
    This abstract class represents the model and forces the
    implementation of certain methods required by the framework.
    """

    @abstractmethod
    def __init__(self, prior, seed = None): 
        """The constructor must be overwritten by a sub-class to initialize the model
        with a given prior.

        The standard behaviour is that concrete model parameters are sampled
        from the provided prior. However, it is alo possible for the constructor
        to provide optional (!) model parameters. In the latter case, the model
        should be initialized by the provided parameters instead from sampling
        from the prior.
    
        Parameters
        ----------
        prior: abcpy.distributions.Distribution
            A prior distribution
        seed: int, optional
            Optional initial seed for the random number generator that can be used in the model. The
            default value is generated randomly.
        """
        raise NotImplemented


    @abstractmethod
    def set_parameters(self, theta):
        """This method properly sets the parameters of the model and must be overwritten by a sub-class.

        Notes
        -----
        Make sure to test whether the provided parameters are
        compatible with the model. Return true if the parameters are accepted
        by the model and false otherwise. This behavior is expected e.g. by
        the inference schemes.

        Parameters
        ----------
        theta:
            An array-like structure containing the p parameter of the model,
            where theta[0] is the first and theta[p-1] is the last parameter.

        Returns
        -------
        boolean
            TRUE if model accepts the provided parameters, FALSE otherwise
        """

        raise NotImplemented


    @abstractmethod
    def sample_from_prior():
        """To be overwritten by any sub-class: should resample the model parameters
        from the prior distribution.
        """        
        raise NotImplemented

    
    @abstractmethod
    def simulate(self, k):
        """To be overwritten by any sub-class: should create k possible outcomes of
        the model using the current model parameters.

        Parameters
        ----------
        k: int
            Number of model outcomes to simulate
            
        Returns
        -------
        Python list 
            An array containing k realizations of the model
        """

        raise NotImplemented
    

    @abstractmethod
    def get_parameters(self):
        """To be overwritten by any sub-class: should extract the parameters from the model. 

        Returns
        -------
        numpy.ndarray 
            An array containing the p parameters of the model
        """

        raise NotImplemented


class Gaussian(Model):
    """This class implements the Gaussian model with unknown mean \
    :math:`\mu` and unknown standard deviation :math:`\sigma`.
    """

    def __init__(self, prior, mu = None, sigma = None, seed=None):
        """
        Parameters
        ----------
        prior: abcpy.distributions.Distribution
            Prior distribution
        mu: float, optional
            Mean of the Gaussian distribution. If the parameters is omitted, sampled
            from the prior.
        sigma: float, optional
            Standard deviation of the Gaussian distribution. If the parameters is omitted, sampled
            from the prior.
        seed: int, optional
            Initial seed. The default value is generated randomly.
        """

        # test prior
        if np.shape(prior.sample(1)) == (1,2):
            self.prior = prior
        else:
            raise ValueError("Prior generates values outside the model "
                             "parameter domain. ")

        # test provided model parameters
        if (mu == None) != (sigma == None):
            raise ValueError("Both or neither of the model parameters have to be provided.")

        # set model parameters directly if specified
        if mu != None and sigma != None:
            if self.set_parameters(np.array([mu, sigma])) == False:
                raise ValueError("The parameter values are out of the model parameter domain.")
        else:
            self.sample_from_prior()


        # set random number generator
        self.rng = np.random.RandomState(seed)

    def set_parameters(self, theta):
        if isinstance(theta, (list,np.ndarray)):
            theta = np.array(theta)
        else:
            raise TypeError('Theta is not of allowed types')

        if theta.shape[0] > 2: return False
        if theta[1] <= 0: return False
        self.mu = theta[0]
        self.sigma = theta[1]
        return True

    def get_parameters(self):
        return np.array([self.mu, self.sigma])

    def sample_from_prior(self):
        sample = self.prior.sample(1).reshape(-1)
        if self.set_parameters(sample) == False:
            raise ValueError("Prior generates values that are out the model parameter domain.")

    def simulate(self, k):
        return list((self.rng.normal(self.mu, self.sigma, k)).reshape(-1))



class Student_t(Model):
    """This class implements the Student_t distribution with unknown mean :math:`\mu` and unknown degrees of freedom.
    """
    def __init__(self, prior, mu=None, df=None, seed=None):
        """
        Parameters
        ----------
        prior: abcpy.distributions.Distribution
            Prior distribution        
        mu: float, optional
            Mean of the Stundent_t distribution. If the parameters is omitted, sampled
            from the prior.
        df: float, optional
            The degrees of freedom of the Student_t distribution. If the parameters is omitted, sampled
            from the prior.
        seed: int, optional
            Initial seed. The default value is generated randomly. 
        """
        assert(not (mu == None) != (df == None))
        self.prior = prior
        if mu == None and df == None:
            self.sample_from_prior()
        else:
            if self.set_parameters(np.array([mu, df])) == False:
                raise ValueError("The parameter values are out of the model parameter domain.")
                
        if not isinstance(prior, Distribution):
            raise TypeError('Prior is not of our defined Prior class type')

        self.rng = np.random.RandomState(seed)
        

    def sample_from_prior(self):
        sample = self.prior.sample(1).reshape(-1)
        if self.set_parameters(sample) == False:
            raise ValueError("Prior generates values that are out the model parameter domain.")         
            
    def simulate(self,k):
        return list((self.rng.standard_t(self.df,k)+self.mu).reshape(-1))

    def get_parameters(self):
        return np.array([self.mu, self.df])

    def set_parameters(self,theta):
        if isinstance(theta, (list,np.ndarray)):
            theta = np.array(theta)
        else:
            raise TypeError('Theta is not of allowed types')

        if theta.shape[0] > 2: return False
        if theta[1] <= 0: return False        
        self.mu = theta[0]
        self.df = theta[1]                      
        return True
        
class MixtureNormal(Model):
    """This class implements the Mixture of multivariate normal ditribution with unknown mean \
    :math:`\mu` described as following,
    :math:`x|\mu \sim 0.5\mathcal{N}(\mu,I_p)+0.5\mathcal{N}(\mu,0.01I_p)`, where :math:`x=(x_1,x_2,\ldots,x_p)` is the 
    dataset simulated from the model and mean is :math:`\mu=(\mu_1,\mu_2,\ldots,\mu_p)`.            """
    def __init__(self, prior, mu, seed = None):
        """
        Parameters
        ----------
        prior: abcpy.distributions.Distribution
            Prior distribution        
        mu: numpy.ndarray or list, optional   
            Mean of the mixture normal. If the parameter is omitted, sampled
            from the prior. 
        seed: int, optional
            Initial seed. The default value is generated randomly. 
        """    
        # Assign prior
        if not isinstance(prior, Distribution):
            raise TypeError('Prior is not of our defined Prior class type') 
        else:
            self.prior = prior
        # Assign parameters
        if isinstance(mu, (list,np.ndarray)):
            if self.set_parameters(mu) == False:
                raise ValueError("The parameter values are out of the model parameter domain.")   
        else:
            raise TypeError('The parameter theta is not of allowed types')
        # Initialize random number generator with provided seed, if None initialize with present time.
        self.rng = np.random.RandomState(seed)
                    
    def sample_from_prior(self):
        sample = self.prior.sample(1).reshape(-1)
        if self.set_parameters(sample) == False:
            raise ValueError("Prior generates values that are out the model parameter domain.")                

    def simulate(self, n_simulate):
        # Generate n_simulate np.ndarray from mixture_normal
        Data_array = [None]*n_simulate
        # Initialize local parameters
        dimension = self.mu.shape[0]
        index_array = self.rng.binomial(1, 0.5, n_simulate)
        for k in range(0,n_simulate):   
            # Initialize the time-series
            index = index_array[k]
            Data = index*self.rng.multivariate_normal(mean = self.mu, cov = np.identity(dimension)) \
                + (1-index)*self.rng.multivariate_normal(mean = self.mu, cov = 0.01*np.identity(dimension))
            Data_array[k] = Data
        # return an array of objects of type Timeseries
        return Data_array
    
    def get_parameters(self):
        return self.mu

    def set_parameters(self, mu):
        if isinstance(mu, (list,np.ndarray)):
            self.mu = np.array(mu)
        else:
            raise TypeError('The parameter value is not of allowed types')
            
        return True        
        
class StochLorenz95(Model):
    """Generates time dependent 'slow' weather variables following forecast model of Wilks [1], 
    a stochastic reparametrization of original Lorenz model Lorenz [2]. 
    
    [1] Wilks, D. S. (2005). Effects of stochastic parametrizations in the lorenz ’96 system. 
    Quarterly Journal of the Royal Meteorological Society, 131(606), 389–407.     

    [2] Lorenz, E. (1995). Predictability: a problem partly solved. In Proceedings of the 
    Seminar on Predictability, volume 1, pages 1–18. European Center on Medium Range
    Weather Forecasting, Europe             
    """    
    
    def __init__(self, prior, theta, initial_state = None, n_timestep = 160, seed = None):
        """
        Parameters
        ----------
        prior: abcpy.distributions.Distribution
            Prior distribution          
        theta: list or numpy.ndarray, optional       
            Closure parameters. If the parameter is omitted, sampled
            from the prior. 
        initial_state: numpy.ndarray, optional
            Initial state value of the time-series, The default value is None, which assumes a previously computed 
            value from a full Lorenz model as the Initial value. 
        n_timestep: int, optional
            Number of timesteps between [0,4], where 4 corresponds to 20 days. The default value is 160.
        seed: int, optional
            Initial seed. The default value is generated randomly.     
        """        
        
        # Assign prior
        if not isinstance(prior, Distribution):
            raise TypeError('Prior is not of our defined Prior class type') 
        else:
            self.prior = prior
        # Assign number of tmestep    
            self.n_timestep = n_timestep
        # Assign initial state    
        if not initial_state == None:
            self.initial_state = initial_state
        else:
            self.initial_state = np.array([6.4558,1.1054,-1.4502,-0.1985,1.1905,2.3887,5.6689,6.7284,0.9301, \
            4.4170,4.0959,2.6830,4.7102,2.5614,-2.9621,2.1459,3.5761,8.1188,3.7343,3.2147,6.3542, \
            4.5297,-0.4911,2.0779,5.4642,1.7152,-1.2533,4.6262,8.5042,0.7487,-1.3709,-0.0520, \
            1.3196,10.0623,-2.4885,-2.1007,3.0754,3.4831,3.5744,6.5790])
        # Assign closure parameters
        if isinstance(theta, (list, np.ndarray)):
            if self.set_parameters(theta) == False:
                raise ValueError("The parameter values are out of the model parameter domain.")
        else:
            raise TypeError('The parameter theta is not of allowed types')
        # Other parameters kept fixed
        self.F = 10
        self.sigma_e = 1
        self.phi = 0.4                               
        # Initialize random number generator with provided seed, if None initialize with present time.
        self.rng = np.random.RandomState(seed)     
                   
    def sample_from_prior(self): 
        sample = self.prior.sample(1).reshape(-1)
        if self.set_parameters(sample) == False:
            raise ValueError("Prior generates values that are out the model parameter domain.")                

    def simulate(self, n_simulate):
        # Generate n_simulate time-series of weather variables satisfying Lorenz 95 equations
        timeseries_array = [None]*n_simulate   
        
        # Initialize timesteps
        time_steps = np.linspace(0, 4, self.n_timestep)
        
        for k in range(0,n_simulate):   
            # Define a parameter object containing parameters which is needed 
            # to evaluate the ODEs
            # Stochastic forcing term 
            eta = self.sigma_e*np.sqrt(1-pow(self.phi,2))*self.rng.normal(0, 1, self.initial_state.shape[0])

            # Initialize the time-series
            timeseries = np.zeros(shape = (self.initial_state.shape[0], self.n_timestep), dtype=np.float)
            timeseries[:,0] = self.initial_state
            # Compute the timeseries for each time steps
            for ind in range(0, self.n_timestep-1):
                # parameters to be supplied to the ODE solver            
                parameter = [eta, self.get_parameters()]
                # Each timestep is computed by using a 4th order Runge-Kutta solver
                x = self._rk4ode(self._l95ode_par, np.array([time_steps[ind], time_steps[ind+1]]), timeseries[:,ind], parameter)
                timeseries[:,ind+1] = x[:,-1]
                # Update stochastic foring term
                eta = self.phi*eta+self.sigma_e*np.sqrt(1-pow(self.phi,2))*self.rng.normal(0, 1)
            timeseries_array[k] = timeseries
        # return an array of objects of type Timeseries
        return timeseries_array
    
    def get_parameters(self):
        return self.theta

    def set_parameters(self, theta):
        if isinstance(theta, (list,np.ndarray)):
            self.theta = np.array(theta)
        else:
            raise TypeError('The parameter value is not of allowed types')
        
        return True
        
    def _l95ode_par(self, t, x, parameter):
        """
        The parameterized two-tier lorenz 95 system defined by a set of symmetic
        ordinary differential equation. This function evaluates the differential 
        equations at a value x of the time-series
 
        Parameters
        ----------
        x: numpy.ndarray of dimension px1
            The value of timeseries where we evaluate the ODE
        parameter: Python list
            The set of parameters needed to evaluate the function
        Returns
        -------
        numpy.ndarray
            Evaluated value of the ode at a fixed timepoint
        """
        # Initialize the array containing the evaluation of ode 
        dx = np.zeros(shape=(x.shape[0],))
        eta = parameter[0]
        theta = parameter[1]
        # Deterministic parameterization for fast weather variables
        # ---------------------------------------------------------
        # assumed to be polynomial, degree of the polynomial same as the 
        # number of columns in closure parameter
        degree = theta.shape[0]
        X = np.ones(shape = (x.shape[0],1))          
        for ind in range(1,degree):
            X = np.column_stack((X,pow(x,ind)))
            
        # deterministic reparametrization term
        # ------------------------------------    
        gu = np.sum(X*theta,1)
      
        # ODE definition of the slow variables
        # ------------------------------------        
        dx[0] = -x[-2]*x[-1] + x[-1]*x[1] - x[0] + self.F - gu[0] + eta[0]
        dx[1] = -x[-1]*x[0] + x[0]*x[2] - x[1] + self.F - gu[1] + eta[1]
        for ind in range(2, x.shape[0] - 2):
            dx[ind] = -x[ind-2]*x[ind-1] + x[ind-1]*x[ind+1] - x[ind] + self.F - gu[ind] + eta[ind]
        dx[-1] = -x[-3]*x[-2]+x[-2]*x[1] - x[-1] + self.F - gu[-1] + eta[-1]

        return dx
        
    def _rk4ode(self, ode, timespan, timeseries_initial, parameter):
        """
        4th order runge-kutta ODE solver.
        
        Parameters
        ----------
        ode: function
            The function defining Ordinary differential equation  
        timespan: numpy.ndarray
            A numpy array containing the timepoints where the ode needs to be solved. 
            The first time point corresponds to the initial value
        timeseries_init: np.ndarray of dimension px1
            Intial value of the time-series, corresponds to the first value of timespan
        parameter: Python list
            The parameters needed to evaluate the ode
        Returns
        -------
        np.ndarray
            Timeseries initiated at timeseries_init and satisfying ode solved by this solver.
        """
        
        timeseries = np.zeros(shape = (timeseries_initial.shape[0], timespan.shape[0]))
        timeseries[:,0] = timeseries_initial

        for ind in range(0, timespan.shape[0]-1):
            time_diff = timespan[ind+1]-timespan[ind] 
            time_mid_point = timespan[ind] + time_diff/2
            k1 = time_diff*ode(timespan[ind], timeseries_initial, parameter)
            k2 = time_diff*ode(time_mid_point, timeseries_initial + k1/2, parameter)
            k3 = time_diff*ode(time_mid_point, timeseries_initial + k2/2, parameter)
            k4 = time_diff*ode(timespan[ind+1], timeseries_initial + k3, parameter)
            timeseries_initial = timeseries_initial + (k1 + 2*k2 + 2*k3 + k4)/6
            timeseries[:,ind+1] = timeseries_initial
        # Return the solved timeseries at the values in timespan    
        return timeseries    
class Ricker(Model):
    """Ecological model that describes the observed size of animal population over time 
    described in [1].
        
    [1] S. N. Wood. Statistical inference for noisy nonlinear ecological 
    dynamic systems. Nature, 466(7310):1102–1104, Aug. 2010.
    """

    def __init__(self, prior, theta=None, n_timestep = 100, seed = None):
        """
        Parameters
        ----------
        prior: abcpy.distributions.Distribution
            Prior distribution          
        theta: list or numpy.ndarray, optional       
            The parameter is a vector consisting of three numbers \
            :math:`\log r` (real number), :math:`\sigma` (positive real number, > 0), :math:`\phi` (positive real number > 0)
            If the parameter is ommitted, sampled from the prior.
        n_timestep: int, optional
            Number of timesteps. The default value is 100.
        seed: int, optional
            Initial seed. The default value is generated randomly.             
        """    
        # Assign prior
        if not isinstance(prior, Distribution):
            raise TypeError('Prior is not of our defined Prior class type') 
        else:
            self.prior = prior
        # Assign number of tmestep    
            self.n_timestep = n_timestep
        # Assign parameters
        if isinstance(theta, (list, np.ndarray)):
            if self.set_parameters(theta) == False:
                raise ValueError("The parameter values are out of the model parameter domain.")   
        # Initialize random number generator with provided seed, if None initialize with random seed.
        self.rng = np.random.RandomState(seed)
                    
    def sample_from_prior(self):
        sample = self.prior.sample(1).reshape(-1)
        if self.set_parameters(sample) == False:
            raise ValueError("Prior generates values that are out the model parameter domain.")                

    def simulate(self, n_simulate):
        timeseries_array = [None]*n_simulate
        # Initialize local parameters
        log_r = self.theta[0]
        sigma = self.theta[1]
        phi = self.theta[2]
        for k in range(0,n_simulate):   
            # Initialize the time-series
            timeseries_obs_size = np.zeros(shape = (self.n_timestep), dtype=np.float)
            timeseries_true_size = np.ones(shape = (self.n_timestep), dtype=np.float)
            for ind in range(1,self.n_timestep-1):
                timeseries_true_size[ind] = np.exp(log_r+np.log(timeseries_true_size[ind-1])-timeseries_true_size[ind-1]+sigma*self.rng.normal(0, 1))
                timeseries_obs_size[ind] = self.rng.poisson(phi*timeseries_true_size[ind])
            timeseries_array[k] = timeseries_obs_size
        # return an array of objects of type Timeseries
        return timeseries_array
    
    def get_parameters(self):
        return self.theta

    def set_parameters(self, theta):
        if isinstance(theta, (list,np.ndarray)):
                theta = np.array(theta)
        else:
            raise TypeError('The parameter value is not of allowed types')            
            
        if theta.shape[0] > 3: return False
        if theta[1] <= 0: return False
        if theta[2] <= 0: return False
        self.theta = theta
        return True



from abc import ABCMeta, abstractmethod
import numpy as np
from sklearn import ensemble



class ModelSelections(metaclass = ABCMeta):
    """This abstract base class defines a model selection rule of how to choose a model from a set of models
    given an observation. 
 
            
    """
    
    @abstractmethod
    def __init__(self, model_array, statistics_calc, backend, seed = None):
        """Constructor that must be overwritten by the sub-class.

        The constructor of a sub-class must accept an array of models to choose the model 
        from, and two non-optional parameters statistics calculator and backend stored in self.statistics_calc 
        and self.backend defining how to calculate sumarry statistics from data and what kind of parallelization
        to use.
        
        Parameters
        ----------
        model_array: list
            A list of models which are of type abcpy.models.Model
        statistics: abcpy.statistics.Statistics
            Statistics object that conforms to the Statistics class.
        backend: abcpy.backends.Backend
            Backend object that conforms to the Backend class.
        seed: integer, optional
            Optional initial seed for the random number generator. The default value is generated randomly.    
        """   
        self.model_array = model_array
        self.statistics_calc = statistics_calc
        self.backend = backend
        self.rng = np.random.RandomState(seed)
        self.reference_table_calculated = None
        

        raise NotImplemented
        
        
    @abstractmethod
    def select_model(self, observations, n_samples = 1000, n_samples_per_param = 100):
        """To be overwritten by any sub-class: returns a model selected by the modelselection
        procedure most suitable to the obersved data set observations. It is assumed that observations is a 
        list of n same type elements(eg., The observations can be a list containing n timeseries, n 
        graphs or n np.ndarray). Further two optional integer arguments n_samples and n_samples_per_param
        is supplied denoting the number of samples in the refernce table and the data points in each 
        simulated data set.
        
        Parameters
        ----------
        observations: python list
            The observed data set.
        n_samples : integer, optional
            Number of samples to generate for reference table.
        n_samples_per_param : integer, optional 
            Number of data points in each simulated data set.  
        Returns
        -------
        abcpy.models.Model
            A model which are of type abcpy.models.Model
            
        """
        
        raise NotImplemented

    @abstractmethod
    def posterior_probability(self, observations):
        """To be overwritten by any sub-class: returns the approximate posterior probability
        of the chosen model given the observed data set observations. It is assumed that observations 
        is a  list of n same type elements(eg., The observations can be a list containing n timeseries, n graphs or n np.ndarray).
        
        Parameters
        ----------
        observations: python list
            The observed data set.
        Returns
        -------
        np.ndarray
            A vector containing the approximate posterior probability of the model chosen.                
        """
        
        raise NotImplemented
        
class RandomForest(ModelSelections):
    """
    This class implements the model selection procedure based on the Random Forest ensemble learner
    as described in Pudlo et. al. [1].
    
    [1] Pudlo, P., Marin, J.-M., Estoup, A., Cornuet, J.-M., Gautier, M. and Robert, C.
    (2016). Reliable ABC model choice via random forests. Bioinformatics, 32 859–866.
    """
    def __init__(self, model_array, statistics_calc, backend, N_tree = 100, n_try_fraction = 0.5, seed = None):
        """        
        Parameters
        ----------
        N_tree : integer, optional
            Number of trees in the random forest. The default value is 100.
        n_try_fraction : float, optional 
            The fraction of number of summary statistics to be considered as the size of 
            the number of covariates randomly sampled at each node by the randomised CART.
            The default value is 0.5.           
        """
        
        self.model_array = model_array
        self.statistics_calc = statistics_calc
        self.backend = backend
        self.rng = np.random.RandomState(seed)
        self.seed = seed 
        self.reference_table_calculated = 0
        self.N_tree = N_tree
        self.n_try_fraction = n_try_fraction

    def select_model(self, observations, n_samples = 1000, n_samples_per_param = 1):
        """        
        Parameters
        ----------
        observations: python list
            The observed data set.
        n_samples : integer, optional
            Number of samples to generate for reference table. The default value is 1000.
        n_samples_per_param : integer, optional 
            Number of data points in each simulated data set. The default value is 1.
        Returns
        -------
        abcpy.models.Model
            A model which are of type abcpy.models.Model
            
        """
        # Creation of reference table
        if self.reference_table_calculated is 0:        
            rc = _RemoteContextForReferenceTable(self.backend, self.model_array, self.statistics_calc, observations, n_samples_per_param)

            # Simulating the data, distance and statistics                
            seed_arr = self.rng.randint(1, n_samples*n_samples, size=n_samples, dtype=np.int32)
            seed_pds = self.backend.parallelize(seed_arr)     

            model_data_pds = self.backend.map(rc._simulate_model_data, seed_pds)
            model_data = self.backend.collect(model_data_pds)
            models, data, statistics = [list(t) for t in zip(*model_data)]
            self.reference_table_models = models
            self.reference_table_data = data
            self.reference_table_statistics = np.concatenate(statistics)
            self.reference_table_calculated = 1

        # Construct a label for the model_array
        label = np.zeros(shape=(len(self.reference_table_models)))
        for ind1 in range(len(self.reference_table_models)):  
            for ind2 in range(len(self.model_array)):
                if self.reference_table_models[ind1] == self.model_array[ind2]:
                    label[ind1] = ind2 

        # Define the classifier 
        classifier = ensemble.RandomForestClassifier(n_estimators = self.N_tree, \
        max_features=int(self.n_try_fraction*self.reference_table_statistics.shape[1]), bootstrap=True, random_state=self.seed)
        classifier.fit(self.reference_table_statistics, label)            

        return(self.model_array[int(classifier.predict(self.statistics_calc.statistics(observations)))])

    def posterior_probability(self, observations, n_samples = 1000, n_samples_per_param = 1):
        """        
        Parameters
        ----------
        observations: python list
            The observed data set.
        n_samples : integer, optional
            Number of samples to generate for reference table. The default value is 1000.
        n_samples_per_param : integer, optional 
            Number of data points in each simulated data set. The default value is 1.
        Returns
        -------
        abcpy.models.Model
            A model which are of type abcpy.models.Model
            
        """        
        # Creation of reference table
        if self.reference_table_calculated is 0:        
            rc = _RemoteContextForReferenceTable(self.backend, self.model_array, self.statistics_calc, observations, n_samples_per_param)

            # Simulating the data, distance and statistics                
            seed_arr = self.rng.randint(1, n_samples*n_samples, size=n_samples, dtype=np.int32)
            seed_pds = self.backend.parallelize(seed_arr)     

            model_data_pds = self.backend.map(rc._simulate_model_data, seed_pds)
            model_data = self.backend.collect(model_data_pds)
            models, data, statistics = [list(t) for t in zip(*model_data)]
            self.reference_table_models = models
            self.reference_table_data = data
            self.reference_table_statistics = np.concatenate(statistics)
            self.reference_table_calculated = 1
        
        # Construct a label for the model_array
        label = np.zeros(shape=(len(self.reference_table_models)))
        for ind1 in range(len(self.reference_table_models)):  
            for ind2 in range(len(self.model_array)):
                if self.reference_table_models[ind1] == self.model_array[ind2]:
                    label[ind1] = ind2 

        # Define the classifier 
        classifier = ensemble.RandomForestClassifier(n_estimators = self.N_tree, \
        max_features=int(self.n_try_fraction*self.reference_table_statistics.shape[1]), bootstrap=True, random_state=self.seed)
        classifier.fit(self.reference_table_statistics, label)  

        pred_error = np.zeros(len(self.reference_table_models),)
        # Compute missclassification error rate
        for ind in range(len(self.reference_table_models)):
            pred_error[ind] = 1 - classifier.predict_proba(self.statistics_calc.statistics(self.reference_table_data[ind]))[0][int(label[ind])] 

        # Estimate a regression function with prediction error as response on summary statitistics of the reference table                
        regressor = ensemble.RandomForestRegressor(n_estimators = self.N_tree)
        regressor.fit(self.reference_table_statistics,pred_error)

        return(1-regressor.predict(self.statistics_calc.statistics(observations)))
        
class _RemoteContextForReferenceTable:
    """
    Contains everything that is sent over the network like broadcast vars and map functions
    """
    
    def __init__(self, backend, model_array, stat_calc, observations, n_samples_per_param):       
        self.model_array = model_array
        self.stat_calc = stat_calc
        self.n_samples_per_param = n_samples_per_param
        
        # these are usually big tables, so we broadcast them to have them once
        # per executor instead of once per task
        self.observations_bds = backend.broadcast(observations)
        
    def _simulate_model_data(self, seed):
        """
        Samples a single model parameter and simulates from it until
        distance between simulated outcome and the observation is
        smaller than eplison.
        
        Parameters
        ----------            
        seed: int
            value of a seed to be used for reseeding
        Returns
        -------
        np.array
            accepted parameter
        """
        #reseed random number genearator
        rng = np.random.RandomState(seed)
        len_model_array = len(self.model_array)
        model = self.model_array[int(sum(np.linspace(0,len_model_array-1,len_model_array)\
        *rng.multinomial(1,(1/len_model_array)*np.ones(len_model_array))))]
        #reseed prior
        model.prior.reseed(seed)
        #sample from prior
        model.sample_from_prior()
        #sample data set, extract statistics and compute distance from y_obs
        y_sim = model.simulate(self.n_samples_per_param)
        statistics = self.stat_calc.statistics(y_sim)
        
        return (model, y_sim, statistics)


from abc import ABCMeta, abstractmethod

import numpy as np
from scipy.stats import multivariate_normal, norm
from scipy.special import gamma


class Distribution(metaclass = ABCMeta):
    """
    This abstract base class represents a distribution. It can be used e.g. as a
    prior for models.

    """


    @abstractmethod
    def set_parameters(self, params):
        """To be overwritten by any sub-class: should set the parameters of the distribution.

        Parameters
        ----------
        theta: list
            Contains all the distributions parameters.
            
        """

        raise NotImplemented


    
    @abstractmethod
    def reseed(self, seed):
        """To be overwritten by any sub-class: reseed the random number generator with provided seed.

        Parameters
        ----------
        seed: integer
            New seed for the random number generator
            
        """

        raise NotImplemented

    
    
    @abstractmethod
    def sample(self, k):
        """To be overwritten by any sub-class: should sample k points from the implemented distribution.

        Parameters
        ----------
        k: integer
            The number of points to be sampled
        Returns
        -------
        numpy.ndarray
            kxp matrix containing k samples of p-dimensional points
            
        """
        
        raise NotImplemented
    

    @abstractmethod
    def pdf(self, x):
        """To be overwritten by any sub-class: calculate the prior-denisty at *x*, where
        *x* is a parameter of dimension p.

        Parameters
        ----------
        x: numpy.ndarray
            A p-dimensional point from the support of the distribution
        Returns
        -------
        float
            The probability density for point x
            
        """
        
        raise NotImplemented
        

class MultiNormal(Distribution):
    """
    This class implements a p-dimensional multivariate Normal distribution.
    """    
    def __init__(self, mean, cov, seed=None):
        """        
        Defines and mean covariance of multivariate normal distribution.
        
        Parameters
        ----------
        mean: numpy.ndarray
            p-dimensional vector containing p means
        cov: numpy.ndarray
            pxp covariance matrix 
        seed: integer
            Initial seed for the random number generator

        """

        self.rng = np.random.RandomState(seed)
        self.mean = mean
        self.cov = cov



    def set_parameters(self, params):      
        self.mean = params[0]
        self.cov = params[1]



    def reseed(self, seed):
        self.rng.seed(seed)
        

        
    def sample(self, k):
        #samples = self.distribution.rvs(k).reshape(k,p)
        samples = self.rng.multivariate_normal(self.mean, self.cov, k)
        return samples
    
    def pdf(self, x):
        return multivariate_normal(self.mean, self.cov).pdf(x)



class Uniform(Distribution):
    """
    This class implements a p-dimensional uniform Prior distribution in a closed interval.
    """    
    def __init__(self, lb : np.ndarray, ub : np.ndarray, seed=None):
        """
        Defines the upper and lower bounds of a p-dimensional uniform Prior distribution in a closed interval.

        Parameters
        ----------
        lb: numpy.ndarray or a list
            Vector containing p lower bounds
        ub: numpy.ndarray or a list
            Vector containing p upper bounds
        seed: integer
            Initial seed for the random number generator

        """

        self.lb, self.ub = self._check_parameters(lb, ub)
        self.rng = np.random.RandomState(seed)

        
    def set_parameters(self, params):
        lb = params[0]
        ub = params[1]
        self.lb, self.ub = self._check_parameters(lb, ub)


    def reseed(self, seed):
        self.rng.seed(seed)
        

    def sample(self, k):
        samples = np.zeros(shape=(k,len(self.lb)))
        for j in range(0,len(self.lb)):
            samples[:,j] = self.rng.uniform(self.lb[j], self.ub[j], k)
        
        return samples
    
        
    def pdf(self,x):
        if np.product(np.greater_equal(x,self.lb)*np.less_equal(x,self.ub)):
            pdf_value = 1/np.product(self.ub-self.lb)
        else:
            pdf_value = 0            
        return pdf_value
        

    def _check_parameters(self, lb, ub):
        new_lb = new_ub = None
        if isinstance(lb, (list,np.ndarray)):
            new_lb = np.array(lb)
        else:
            raise TypeError('The lower bound is not of allowed types')

        if isinstance(ub, (list,np.ndarray)):
            new_ub = np.array(ub)
        else:
            raise TypeError('The upper bound is not of allowed types')
                
        if new_lb.shape != new_ub.shape:
            raise BaseException('Dimension of lower bound and upper bound is not same.')

        return (new_lb, new_ub)
        



class MultiStudentT(Distribution):
    """
    This class implements a p-dimensional multivariate Student T distribution.
    
    """
    
    def __init__(self, mean, cov, df, seed=None):
        """Defines the mean, co-variance and degrees of freedom a p-dimensional multivariate Student T distribution.

        Parameters
        ----------
        mean: numpy.ndarray
            Vector containing p means, one for every dimension        
        cov: numpy.ndarray
            pxp matrix containing the co-variance matrix        
        df: np.uint
            Degrees of freedom
 
        """

        MultiStudentT._check_parameters(mean, cov, df)
        
        self.mean = mean
        self.cov = cov
        self.df = df
        self.rng = np.random.RandomState(seed)


        
    def set_parameters(self, params):      
        mean = params[0]
        cov = params[1]
        df = self.df
        
        if len(params) == 3:
            df = params[2]
        
        try:
            MultiStudentT._check_parameters(mean, cov, df)
        except TypeError or ValueError:
            return False

        self.mean = mean
        self.cov = cov
    

        
    def reseed(self, seed):
        self.rng.seed(seed)


    
    def sample(self, k):
        p = len(self.mean)
        if self.df == np.inf:
            chisq = 1.0
        else:
            chisq = self.rng.chisquare(self.df, k) / self.df
            chisq = chisq.reshape(-1,1).repeat(p, axis=1)
        mvn = self.rng.multivariate_normal(np.zeros(p), self.cov, k)
        return self.mean + np.divide(mvn, np.sqrt(chisq))

    
    
    def pdf(self, x):
        m = self.mean
        cov = self.cov
        v = self.df
        p = len(self.mean)
        
        numerator = gamma((v + p) / 2)
        denominator = gamma(v/2) * pow(v * np.pi, p/2.) * np.sqrt(abs(np.linalg.det(cov)))
        normalizing_const = numerator / denominator
        tmp = 1 + 1/v * np.dot(np.dot(np.transpose(x - m), np.linalg.inv(cov)), (x-m))
        density = normalizing_const * pow(tmp, -((v+p)/2.))
        return density
        

    
    @classmethod
    def _check_parameters(cls, mean, cov, df):
        """
        Private function that checks whether the parameters make sense for the distribution.
        """

        if not isinstance(mean, np.ndarray):
            raise TypeError("Parameter mean is not of type numpy.array.")

        if not isinstance(cov, np.ndarray):
            raise TypeError("Parameter cov is not of type numpy.array.")

        if mean.ndim != 1:
            raise ValueError("Parameter mean has dimension larger than 1.")

        if cov.ndim != 2:
            raise ValueError("Parameter cov is not of dimensionality 2.")

        if df <= 0:
            raise ValueError("Parameter df is smaller than 0.")
            
class Normal(Distribution):
    """
    This class implements a 1-dimensional Normal distribution.
    """    
    def __init__(self, mean, var, seed=None):
        """        
        Defines and mean and variance of normal distribution.
        
        Parameters
        ----------
        mean: numpy.ndarray
            mean
        var: numpy.ndarray
            variance
        seed: integer
            Initial seed for the random number generator

        """

        self.rng = np.random.RandomState(seed)
        self.mean = mean
        self.var = var



    def set_parameters(self, params):      
        self.mean = params[0]
        self.var = params[1]



    def reseed(self, seed):
        self.rng.seed(seed)
        

        
    def sample(self, k):
        #samples = self.distribution.rvs(k).reshape(k,p)
        samples = self.rng.normal(self.mean, self.var, k)
        return samples.reshape((k,1))
    
    def pdf(self, x):
        return norm(self.mean, self.var).pdf(x)
            
            



import unittest
import numpy as np

from abcpy.models import Gaussian
from abcpy.distributions import Uniform
from abcpy.statistics import Identity
from abcpy.approx_lhd import PenLogReg, SynLiklihood


class PenLogRegTests(unittest.TestCase):
    def setUp(self):
        self.prior = Uniform([-5.0, 5.0], [5.0, 10.0], seed=1)
        self.model = Gaussian(self.prior, mu=1.1, sigma=1.0, seed=1)
        self.stat_calc = Identity(degree = 2, cross = 0)
        self.likfun = PenLogReg(self.stat_calc, self.model, n_simulate = 100, n_folds = 10, max_iter = 100000, seed = 1)

    def test_likelihood(self):
        
        #Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.likfun.likelihood, 3.4, [2,1])
        self.assertRaises(TypeError, self.likfun.likelihood, [2,4], 3.4)

        # create observed data
        y_obs = self.model.simulate(1)
        # create fake simulated data
        self.model.set_parameters(np.array([1.1,1.0]))
        y_sim = self.model.simulate(100)
        comp_likelihood = self.likfun.likelihood(y_obs, y_sim)
        expected_likelihood = 4.3996556327224594
        # This checks whether it computes a correct value and dimension is right
        self.assertLess(comp_likelihood - expected_likelihood, 10e-2)
        
class SynLiklihoodTests(unittest.TestCase):
    def setUp(self):
        self.prior = Uniform([-5.0, 5.0], [5.0, 10.0], seed=1)
        self.model = Gaussian(self.prior, mu=1.1, sigma=1.0, seed = 1)
        self.stat_calc = Identity(degree = 2, cross = 0)
        self.likfun = SynLiklihood(self.stat_calc) 


    def test_likelihood(self):
        #Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.likfun.likelihood, 3.4, [2,1])
        self.assertRaises(TypeError, self.likfun.likelihood, [2,4], 3.4)
               
        # create observed data
        y_obs = self.model.simulate(1)
        # create fake simulated data
        self.model.set_parameters(np.array([1.1,1.0]))
        y_sim = self.model.simulate(100)
        # calculate the statistics of the observed data
        comp_likelihood = self.likfun.likelihood(y_obs, y_sim)
        expected_likelihood = 0.00924953470649
        # This checks whether it computes a correct value and dimension is right
        self.assertLess(comp_likelihood - expected_likelihood, 10e-2)
        



import unittest
import numpy as np
from abcpy.distributions import Uniform
from abcpy.models import Gaussian
from abcpy.statistics import Identity
from abcpy.backends import BackendDummy as Backend
from abcpy.summaryselections import Semiautomatic



class SemiautomaticTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree = 1, cross = 0)
        
        # define prior and model
        prior = Uniform([150, 5],[200, 25])
        self.model = Gaussian(prior, seed = 1)

        # define backend
        self.backend = Backend()

        # define statistics
        self.statistics_cal = Identity(degree = 2, cross = False)
        
        #Initialize summaryselection
        self.summaryselection = Semiautomatic(self.model, self.statistics_cal, self.backend, n_samples = 1000, seed = 1)

        
    def test_transformation(self):
        #Transform statistics extraction
        self.statistics_cal.statistics = lambda x, f2=self.summaryselection.transformation, f1=self.statistics_cal.statistics: f2(f1(x))
        y_obs = self.model.simulate(10)
        extracted_statistics_10 = self.statistics_cal.statistics(y_obs)
        self.assertEqual(np.shape(extracted_statistics_10), (10,2))
        y_obs = self.model.simulate(1)
        extracted_statistics_1 = self.statistics_cal.statistics(y_obs)
        self.assertLess(extracted_statistics_1[0,0] - 111.012664458, 10e-2)
        self.assertLess(extracted_statistics_1[0,1] - (-63.224510811), 10e-2)
        
if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod

class Backend(metaclass = ABCMeta):
    """
    This is the base class for every parallelization backend. It essentially
    resembles the map/reduce API from Spark.

    An idea for the future is to implement a MPI version of the backend with the
    hope to be more complient with standard HPC infrastructure and a potential
    speed-up.

    """

    @abstractmethod
    def parallelize(self, list):
        """
        This method distributes the list on the available workers and returns a
        reference object.

        The list should be split into number of workers many parts. Each
        part should then be sent to a separate worker node.

        Parameters
        ----------
        list: Python list
            the list that should get distributed on the worker nodes
        Returns
        -------
        PDS class (parallel data set)
            A reference object that represents the parallelized list
        """
        
        raise NotImplemented


    @abstractmethod
    def broadcast(self, object):
        """
        Send object to all worker nodes without splitting it up.

        Parameters
        ----------
        object: Python object
            An abitrary object that should be available on all workers

        Returns
        -------
        BDS class (broadcast data set)
            A reference to the broadcasted object
        """
        
        raise NotImplemented


    @abstractmethod
    def map(self, func, pds):
        """
        A distributed implementation of map that works on parallel data sets (PDS).

        On every element of pds the function func is called.

        Parameters
        ----------
        func: Python func
            A function that can be applied to every element of the pds
        pds: PDS class
            A parallel data set to which func should be applied
        
        Returns
        -------
        PDS class
            a new parallel data set that contains the result of the map
        """
        
        raise NotImplemented


    @abstractmethod
    def collect(self, pds):
        """
        Gather the pds from all the workers, send it to the master and return it as a standard Python list.

        Parameters
        ----------
        pds: PDS class
            a parallel data set
            
        Returns
        -------
        Python list
            all elements of pds as a list
        """
        
        raise NotImplemented

    
class PDS:
    """
    The reference class for parallel data sets (PDS).
    """

    @abstractmethod
    def __init__(self):
        raise NotImplemented


class BDS:
    """
    The reference class for broadcast data set (BDS).
    """
    
    @abstractmethod
    def __init__(self):
        raise NotImplemented


    @abstractmethod
    def value(self):
        """
        This method should return the actual object that the broadcast data set represents. 
        """
        raise NotImplemented


class BackendDummy(Backend):
    """
    This is a dummy parallelization backend, meaning it doesn't parallelize
    anything. It is mainly implemented for testing purpose.

    """
    
    def __init__(self):
        pass

    
    def parallelize(self, python_list):
        """
        This actually does nothing: it just wraps the Python list into dummy pds (PDSDummy).

        Parameters
        ----------
        python_list: Python list
        Returns
        -------        
        PDSDummy (parallel data set)
        """
        
        return PDSDummy(python_list)

    
    def broadcast(self, object):
        """
        This actually does nothing: it just wraps the object into BDSDummy.

        Parameters
        ----------
        object: Python object
        
        Returns
        -------        
        BDSDummy class
        """
        
        return BDSDummy(object)

    
    def map(self, func, pds):
        """
        This is a wrapper for the Python internal map function.

        Parameters
        ----------
        func: Python func
            A function that can be applied to every element of the pds
        pds: PDSDummy class
            A pseudo-parallel data set to which func should be applied
        
        Returns
        -------
        PDSDummy class
            a new pseudo-parallel data set that contains the result of the map
        """
        
        result_map = map(func, pds.python_list)
        result_pds = PDSDummy(list(result_map))
        return result_pds

    
    def collect(self, pds):
        """
        Returns the Python list stored in PDSDummy

        Parameters
        ----------
        pds: PDSDummy class
            a pseudo-parallel data set
        Returns
        -------
        Python list
            all elements of pds as a list
        """
        
        return pds.python_list

    

class PDSDummy(PDS):
    """
    This is a wrapper for a Python list to fake parallelization.
    """
    
    def __init__(self, python_list):
        self.python_list = python_list

        

class BDSDummy(BDS):
    """
    This is a wrapper for a Python object to fake parallelization.
    """
    
    def __init__(self, object):
        self.object = object

        
    def value(self):
        return self.object

    


class BackendSpark(Backend):
    """
    A parallelization backend for Apache Spark. It is essetially a wrapper for
    the required Spark functionality.
    """
    
    def __init__(self, sparkContext, parallelism=4):
        """
        Initialize the backend with an existing and configured SparkContext.

        Parameters
        ----------
        sparkContext: pyspark.SparkContext
            an existing and fully configured PySpark context
        parallelism: int
            defines on how many workers a distributed dataset can be distributed
        """
        self.sc = sparkContext
        self.parallelism = parallelism


    def parallelize(self, python_list):
        """
        This is a wrapper of pyspark.SparkContext.parallelize().

        Parameters
        ----------
        list: Python list
            list that is distributed on the workers
        
        Returns
        -------
        PDSSpark class (parallel data set)
            A reference object that represents the parallelized list
        """
        
        rdd = self.sc.parallelize(python_list, self.parallelism)
        pds = PDSSpark(rdd)
        return pds


    def broadcast(self, object):
        """
        This is a wrapper for pyspark.SparkContext.broadcast().

        Parameters
        ----------
        object: Python object
            An abitrary object that should be available on all workers
        Returns
        -------
        BDSSpark class (broadcast data set)
            A reference to the broadcasted object
        """
        
        bcv = self.sc.broadcast(object)
        bds = BDSSpark(bcv)
        return bds


    def map(self, func, pds):
        """
        This is a wrapper for pyspark.rdd.map()

        Parameters
        ----------
        func: Python func
            A function that can be applied to every element of the pds
        pds: PDSSpark class
            A parallel data set to which func should be applied
        Returns
        -------
        PDSSpark class
            a new parallel data set that contains the result of the map
        """
        
        rdd = pds.rdd.map(func)
        new_pds = PDSSpark(rdd)
        return new_pds


    def collect(self, pds):
        """
        A wrapper for pyspark.rdd.collect()

        Parameters
        ----------
        pds: PDSSpark class
            a parallel data set
        Returns
        -------
        Python list
            all elements of pds as a list
        """
        
        python_list = pds.rdd.collect()
        return python_list

    
    
class PDSSpark(PDS):
    """
    This is a wrapper for Apache Spark RDDs.
    """
    
    def __init__(self, rdd):
        """
        Returns
        -------
        rdd: pyspark.rdd
            initialize with an Spark RDD
        """
        
        self.rdd = rdd



class BDSSpark(BDS):
    """
    This is a wrapper for Apache Spark Broadcast variables.
    """
    
    def __init__(self, bcv):
        """
        Parameters
        ----------
        bcv: pyspark.broadcast.Broadcast
            Initialize with a Spark broadcast variable
        """
        
        self.bcv = bcv


    def value(self):
        """
        Returns
        -------
        object
            returns the referenced object that was broadcasted.
        """
        
        return self.bcv.value



from abc import ABCMeta, abstractmethod
import numpy as np
from sklearn import linear_model

class Summaryselections(metaclass = ABCMeta):
    """This abstract base class defines a way to choose the summary statistics.
    """
    
    @abstractmethod
    def __init__(self, model, statistics_calc, backend, n_samples = 1000, seed = None):
        """The constructor of a sub-class must accept a non-optional model, statistics calculator and 
        backend which are stored to self.model, self.statistics_calc and self.backend. Further it 
        accepts two optional parameters n_samples and seed defining the number of simulated dataset
        used for the pilot to decide the summary statistics and the integer to initialize the random 
        number generator.
    
        Parameters
        ----------
        model: abcpy.models.Model
            Model object that conforms to the Model class.
        statistics_cal: abcpy.statistics.Statistics
            Statistics object that conforms to the Statistics class.
        backend: abcpy.backends.Backend
            Backend object that conforms to the Backend class.
        n_samples: int, optional
            The number of (parameter, simulated data) tuple generated to learn the summary statistics in pilot step. 
            The default value is 1000. 
        seed: integer, optional
            Optional initial seed for the random number generator. The default value is generated randomly.    
        """
        raise NotImplemented
        
        
    @abstractmethod
    def transformation(self, statistics):
        raise NotImplemented


class Semiautomatic(Summaryselections):
    """This class implements the semi auomatic summary statistics choice described in Fearnhead and Prangle [1].
    
    [1] Fearnhead P., Prangle D. 2012. Constructing summary statistics for approximate
    Bayesian computation: semi-automatic approximate Bayesian computation. J. Roy. Stat. Soc. B 74:419–474.    
    """

    def __init__(self, model, statistics_calc, backend, n_samples = 1000, seed = None):
        self.model = model
        self.statistics_calc = statistics_calc
        self.backend = backend
        self.rng = np.random.RandomState(seed)
        self.model.prior.reseed(self.rng.randint(np.iinfo(np.uint32).max, dtype=np.uint32)) 
        
        rc = _RemoteContextSemiautomatic(self.backend, self.model, self.statistics_calc)
        
        # main algorithm                 
        seed_arr = self.rng.randint(1, n_samples*n_samples, size=n_samples, dtype=np.int32)
        seed_pds = self.backend.parallelize(seed_arr)     

        sample_parameters_statistics_pds = self.backend.map(rc._sample_parameter_statistics, seed_pds)
        sample_parameters_and_statistics = self.backend.collect(sample_parameters_statistics_pds)
        sample_parameters, sample_statistics = [list(t) for t in zip(*sample_parameters_and_statistics)]
        sample_parameters = np.array(sample_parameters)
        sample_statistics = np.concatenate(sample_statistics)
        
        self.coefficients_learnt = np.zeros(shape=(sample_parameters.shape[1],sample_statistics.shape[1]))
        regr = linear_model.LinearRegression(fit_intercept=True)
        for ind in range(sample_parameters.shape[1]):
            regr.fit(sample_statistics, sample_parameters[:,ind]) 
            self.coefficients_learnt[:,ind] = regr.coef_ 
        
    def transformation(self, statistics):
        if not statistics.shape[1] == self.coefficients_learnt.shape[0]:    
            raise ValueError('Mismatch in dimension of summary statistics')
        return np.dot(statistics,self.coefficients_learnt)
        
        
class _RemoteContextSemiautomatic:
    """
    Contains everything that is sent over the network like broadcast vars and map functions
    """
    
    def __init__(self, backend, model, stat_calc):
        self.model = model
        self.stat_calc = stat_calc

        
    def _sample_parameter_statistics(self, seed):
        """
        Samples a single model parameter and simulates from it until
        distance between simulated outcome and the observation is
        smaller than eplison.
        
        Parameters
        ----------            
        seed: int
            value of a seed to be used for reseeding
        Returns
        -------
        np.array
            accepted parameter
        """
        self.model.prior.reseed(seed)
        self.model.sample_from_prior()
        
        return (self.model.get_parameters(), self.stat_calc.statistics(self.model.simulate(1)))     
        
        
        